<?
/* --------------------------------------------------------------------
   Copyright (c) 2004-2012 Daniel Pedemont (dan@pedemont.net)
   -------------------------------------------------------------------- */
/**
 * the two main stream classes
 *
 * the thread class is a fork wrapper which also sets the internal pid numbers for the base class
 *
 * the stream class is the main stream plugin wrapper with utility and helper functions
 * it extends the thread class which provides the sonic_thread_run() function which creates the
 * forked sonic_stream_thread_main() fork wrapper alternative to standard sonic_stream_main().
 * 
 * sonic_stream_main() is provided by the actual plugin which uses said main two classes.
 *
 * @link http://dev.pedemont.com/sonic/
 * @link http://dev.pedemont.com/sonic/downloads/src/lib/sonic.stream.inc
 * @package sonic
 * @version 0.1.15
 * @abstract SONIC_thread and SONIC_stream classes
*/

define('SONIC_STREAM_VERSION', '0.1.17');

//NB, function names in rotate mode have no protection,
//    so dont use generic function names for clash safety

class SONIC_thread
{
  /**
   * parent var for pid of fork
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @see SONIC_thread::sonic_thread_run()
   * @var integer
  */
  var $pid = 0;
  /**
   * child var for pid of fork
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @see SONIC_thread::sonic_thread_run()
   * @var integer
  */
  var $pid_child = 0;

  /**
   * ob_start()
   *
   * if set as 'buffer' ob flush and restart post fork.
   * default is NULL, setup via extended stream constructor
   * over manual override by daemon afterwards.
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @see SONIC_thread::sonic_thread_run()
   * @see ob_start()
   * @see ob_get_contents()
   * @see ob_end_clean()
   * @var string
  */
  var $thread_out = NULL;
  /**
   * log only
   *
   * if output is set as 'buffer' with loggin enabled,
   * then surpress output to standard out aka STDOUT.
   * default is NULL, setup via extended stream constructor
   * over manual override by daemon afterwards.
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @see SONIC_stream::sonic_stream_buffer()
   * @see ob_get_contents()
   * @var boolean
  */
  var $thread_buffer_quiet = NULL;
  /**
   * log file
   *
   * output log file when in threaded mode.
   * when in rotate mode, daemon will catch
   * ob contents and log instead of child.
   * default is NULL, setup via extended stream constructor
   * over manual override by daemon afterwards.
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @see SONIC_thread::sonic_thread_run()
   * @var string
  */
  var $thread_log = NULL;

  /**
   * fork entire program Memory and Execution space aka M.E.
   *
   * forks execution at point, assign pid numbers of both processes 
   * and returns
   *
   * @return boolean pid number if in parents M.E. space, false in childs M.E. space, or dies on failure
  */
  function sonic_thread_run()
  {
    //posix isnt needed but used alot
//    if (extension_loaded('pcntl') && extension_loaded('posix'));
    if (SONICD_PCNTL);
    else
      return false;

    $pid = pcntl_fork();

    if ($pid == -1)
    {
      die('! Error: could not fork thread'."\n");
    }
    else if ($pid)
    {
      $this->pid = posix_getpid();
      $this->pid_child = $pid;
      return $this->pid; // we are the parent
    }
    else
    {
      $this->pid = posix_getpid();
      $this->pid_child = 0;
      usleep(1); //give parent a chance in the race

      if ($this->thread_out == 'buffer')
      {
        $out = ob_get_contents(); //clear prefork
        ob_end_clean();
        ob_start();
      }
   
      return false; // we are the child
    }
  }
}

class SONIC_stream extends SONIC_thread
{
  /**
   * unique random family key for all spawned children of same service class set up the daemon
   *
   * @see SONIC_stream::sonic_stream_main()
   * @var string;
  */
  var $key = NULL;

  /**
   * marks if stream is currently busy or idle
   *
   * @see SONIC_stream::sonic_stream_main()
   * @var boolean
  */
  var $idle = true;
  /**
   * stream config array of key value pairs
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @var array
  */
  var $config = NULL;
  /**
   * verbosity level
   *
   * manually overriden by daemon if not supplied by stream setup
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @var integer
  */
  var $verbose = 0;

  //child core listen helper
  /**
   * socket resource var
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @var integer
  */
  var $sock = NULL;
  /**
   * name of service using stream
   *
   * has preg_replace('/[^az-_0-9]/', '_', NAME); applied for 
   * indexing reasons
   *
   * @see sonic-daemon.php::sonic_daemon_main()
   * @var string
  */
  var $name = NULL; 

  //child ext/util helpers
  /**
   * true if PHP environment has posix and control extensions loaded
   *
   * @see SONIC_stream constructor
   * @var boolean
  */
  var $pcntl_loaded = false;
  /**
   * rotate mode exit (return to daemon)
   *
   * internal flag switched upon shutdown when in rotate type mode
   * should be automagically set my sonic_stream_main() aka the plugin.
   * usually on pcntl signals SIGTERM, SIGQUIT and SIGINT
   *
   * @see SONIC_stream::sonic_stream_main()
   * @var boolean
  */
  var $rotate_shutdown = false;
  /**
   * peername
   *
   * cache of socket var connected client peername
   *
   * @see SONIC_stream::sonic_socket_getpeername()
   * @var string
  */
  var $sock_peername = NULL;
  /**
   * peername (via stunnel)
   *
   * cache of stunnels connected client peername matching local client peername of sock var
   *
   * @see SONIC_stream::sonic_stunnel_getpeername()
   * @var string
  */
  var $stunnel_peername = NULL;

  /**
   * stream constructor
   *
   * loads internal variables with supplied configuration key paired array.
   * usually non service specific, but stream environment related.
   *
   * @param array $config option flags, (string) name, (resource) socket, (integer) verbose, (string) thread_out, (boolean) thread_buffer_quiet, (string) thread_log
  */
  function SONIC_stream($config = NULL)
  {
    $this->config = $config;

    if ($this->config['name'])
    {
      //make name generically index friendly
      $this->config['name'] = preg_replace('/[^a-z0-9\-_]/', '_', $this->config['name']);
      $this->name = &$config['name'];
    }

//key
    if ($this->config['key'])
      $this->key = &$config['key'];

    //TODO?: is this really need/want/ed, safe much? config set static key?
    if ($this->config['stream_config']['key'])
      $this->key = &$config['stream_config']['key'];

//socket
    if ($this->config['socket'])
      $this->sock = &$config['socket'];

//verbose
    if ($this->config['verbose'])
      $this->verbose = &$config['verbose'];

    if ($this->config['stream_config']['verbose'])
      $this->verbose = &$config['stream_config']['verbose'];

//thread_
    if ($this->config['thread_out'])
      $this->thread_out = &$config['thread_out'];

    if ($this->config['thread_buffer_quiet'])
      $this->thread_buffer_quiet = &$config['thread_buffer_quiet'];

    if ($this->config['thread_log'])
    {
      $_thread_log = &$config['thread_log'];

      if ($_thread_log{strlen($_thread_log) - 1} == DIRECTORY_SEPARATOR)
        $_thread_log = substr($_thread_log, 0, -1);

      if (file_exists($_thread_log))
      {
        //echo append notice
        if (is_dir($_thread_log))
        { 
          $_thread_log .= '/sonic-daemon-'.$this->name.'-'.time().'.log';

          echo '$ Notice: output logfile -> \''.$_thread_log.'\''."\n";
        }
      }

      //write permission test or disable with notice
      if (sonic_file_mode_contents($_thread_log, '', 'a'))
      {
        //&$config['thread_log'];
        $this->thread_log = &$_thread_log;
      }
      else
      {
        echo '! '.sonic_stream_prelog()
            .' [SONIC_STREAM] Warning: unable to write logfile -> \''.$_thread_log.'\''."\n";
      }
    }

//    $this->pcntl_loaded = (extension_loaded('pcntl') && extension_loaded('posix'));
    $this->pcntl_loaded = SONICD_PCNTL;
  }

  /**
   * idle check
   *
   * checks if stream has marked itself as busy or not
   *
   * NB, currently not in use by daemon/IPC
   *
   * @return boolean boolean value of internal variable 'idle'
  */
  function sonic_stream_idle() { return $this->idle; }

  //function sonic_stream_main() {} //overloaded by child object

  /**
   * threaded stream main (wrapper)
   *
   * call this function instead of calling sonic_stream_main() directly
   * to fork the process and set internal pid numbers.
   *
   * @return mixed will exit with return value from sonic_stream_main()
  */
  function sonic_stream_thread_main()
  {
    if (parent::sonic_thread_run())
      return true;

    $this->sonic_stream_main();

    if ($this->thread_out == 'buffer')
    {
      //fork buff safety clear
      $out = ob_get_contents();
      ob_end_clean();
      if (!$this->thread_buffer_quiet)
        echo $out;

      if ($this->thread_log && $out != '')
      { 
        sonic_file_mode_contents($this->thread_log, $out, 'a');

        if ($verbose > 9998)
          echo '= '.sonic_stream_prelog()
              .' [SONIC_STREAM] lock safe write to log -> \''.$this->thread_log.'\''."\n";
      }

      unset($out);

      if ($this->verbose > 998)
      { 
        echo ':'; //buffer cycle tick
        flush();
      }

//      ob_start(); //not really needed anymore
    }

    exit();
  }

  /**
   * flush()
   *
   * buffer-flush for output mode 'flush' or 'buffer'
   * NB, will also log set to var thread_log and in
   * output mode set as 'buffer'. ie. also used 
   * for threaded child log write (as log write is
   * handled by the server daemon in rotate mode)
   *
   * @see ob_start()
   * @see ob_end_clean()
   * @see flush()
   * @see sonic.inc::sonic_file_mode_contents()
  */
  function sonic_stream_buffer()
  {
    if ($this->thread_out == 'buffer');
    else
    {
      flush();
      return;
    }

    //fork buff safety clear
    $out = ob_get_contents();
    ob_end_clean();
    if (!$this->thread_buffer_quiet)
      echo $out;

    if ($this->thread_log && $out != '')
    { 
      sonic_file_mode_contents($this->thread_log, $out, 'a');

      if ($verbose > 9998)
        echo '= '.sonic_stream_prelog()
            .' [SONIC_STREAM] lock safe write to log -> \''.$this->thread_log.'\''."\n";
    }

    unset($out);

    ob_start();
  }

  //core utility
  /**
   * log output unifier
   *
   * returns a uniform prefix for log/output entries
   *
   * @return string string usually with pid/process id and timestamp
  */
  function sonic_stream_prelog()
  {
    return '['.($this->pcntl_loaded ? posix_getpid() : getmypid()).'] '
          .sonic_datef()
          .' ['.$this->name.']';
  }

  /**
   * socket_getpeername()
   *
   * @param string &$peername host:port format of client peername on socket
   * @param array $config third party support (library mode) override, array of type see stream constructor
   * @param resource $sock third party support (library mode) override, see PHP sockets
   * @return boolean boolean see socket_getpeername()
  */
  function sonic_stream_socket_getpeername(&$peername, $config = NULL, $sock = NULL)
  {
    $peername = NULL;

    if ($config == NULL && $sock == NULL) //inuse, not library mode
    {
      if ($this->sock_peername)
      {
        $peername = $this->sock_peername;
        return true;
      }
    }

    if ($config) {}               //3rd party compat
    else if ($this->config)       //sonic daemon inherit preload
    {
      $config = $this->config;
    }
    else
      return false;

    if ($sock) {}                 //3rd party compat
    else if ($this->sock)         //sonic daemon inherit preload
    {
      $sock = &$this->sock;
    }
    else if ($config['socket'])   //despro search
    {
      $sock = &$config['socket'];
    }
    else
      return false;

    if ($sgpn_ret = socket_getpeername($sock, $host, $port));
    else
      return false;

    $peername = "$host:$port";
    return $sgpn_ret;
  }

  /**
   * _polling_ signal handler for socket_accept()
   *
   * socket_accept by nature will hang process flow and ignore
   * any other signals accept SIGALRM
   *
   * using this will allow other signals to be processed without
   * interruption
   *
   * @param integer $sig posix process signal
  */
  function _sonic_stream_socket_accept_sighandler($sig = NULL)
  {
//echo "_sonic_stream_socket_accept_sighandler()\n";
    //WARNING (see SIGALRM hack)
    //this should work, but socket_accept seems to have
    //its own handlers or ignores set handlers when in progress
    switch ($sig)
    { 
      case SIGALRM:
        socket_clear_error($this->sock);
//        echo "debug: sonic_stream_socket_accept poll hack\n";
    }
  }

  /**
   * socket_accept()
   *
   * polls (if pcntl is supported and enabled) socket_accept() and sets internal variable 'peername' 
   * via get_peername (stunnel via config supported)
   *
   * @param string &$stat_msg return status information
   * @param array $config third party support (library mode) override, array of type see stream constructor
   * @param resource $sock third party support (library mode) override, see PHP sockets
   * @param boolean $pcntl_loaded third party support (library mode) override
   * @return boolean boolean see socket_getpeername()
  */
  function sonic_stream_socket_accept(&$stat_msg, $config = NULL, $sock = NULL, $pcntl_loaded = NULL)
  {
    $stat_msg = 'OK';

    if ($this->config == NULL && $this->$sock == NULL)
      $_self_inuse = false;
    else //inuse, not library mode
      $_self_inuse = true;

    if ($config) {}                       //3rd party compat
    else if ($this->config)               //sonic daemon inherit preload
    {
//TODO: more info about stale cached peernames?
      $this->sock_peername = NULL;
      $this->stunnel_peername = NULL;

      $config = $this->config;
    }
    else
      return false;

    if ($sock) {}                         //3rd party compat
    else if ($this->sock)                 //sonic daemon inherit preload
    {
      $sock = &$this->sock;
    }
    else if ($config['socket'])           //despro search
    {
      $sock = &$config['socket'];
    }
    else
      return false;

    if ($pcntl_loaded != NULL) {}         //3rd party compat
    else if ($this->pcntl_loaded != NULL) //sonic daemon inherit preload
    {
      $pcntl_loaded = $this->pcntl_loaded;
    }
    else
      $pcntl_loaded = false;

    $stream_config = $config['stream_config'];

    $sock_peername = '';
    if ($config['socket_type'] == SOCK_STREAM)
    {
      //socket accept supresses signal handlers except sigalarm,
      //will need to break and return until catch so our
      //sig handlers actually get a chance to run
      $_si_accepted = false;

      if ($pcntl_loaded)
        pcntl_signal(SIGALRM, array(&$this, '_sonic_stream_socket_accept_sighandler'));

      do
      {
        if ($pcntl_loaded)
          pcntl_alarm(1);

        if (($_sock = @socket_accept($sock)) == false) //0 ok but nothing
        {
//          var_dump(socket_last_error()); //interrupt can only be (for us here) sigalrm
          if ($pcntl_loaded && socket_last_error() == 4);    //Interrupted system call
          else
          {
            $stat_msg = socket_strerror(socket_last_error($sock));
            return false;
          }
        }
        else
        {
          $_st_filter_time = time();

          if ($this->pcntl_loaded)
            pcntl_alarm(0);

          $_si_accepted = true;

          if (socket_getpeername($_sock, $sock_phost, $sock_pport))
            $sock_peername = "$sock_phost:$sock_pport";
          else
          {
            socket_close($_sock);
            $stat_msg = socket_strerror(socket_last_error($_sock));
            return false;
          }
        }

        if ($_si_accepted)
          break;

        if ($this->rotate_shutdown)
          break;

      } while (true);
      unset($_si_accepted);

      if ($_sock && $sock_peername)
        $stat_msg = 'connected peername -> '.$sock_peername;
      else
        return false;

      //post-accept, stunnel real peername support
      if ($stream_config['stunnel_log'] OR $stream_config['stunnel_force'])
      {
        $st_filter = array();
        $st_filter['time'] = $_st_filter_time;
        $st_filter['server'] = $sock_peername;

        if (SONIC_stream::sonic_stream_stunnel_getpeername($st_peername, $st_filter, $config, $_sock))
        {
          $stat_msg = 'stunnel -> '.basename($stream_config['stunnel_log']).' -> '.$st_peername;
//should already be assigned by stunnel_peername when inuse mode
//          if ($_self_inuse) //inuse, not library mode
//            $this->stunnel_peername = &$st_peername;
        }
        else if ($stream_config['stunnel_force'])
        {
          socket_close($_sock);
          $stat_msg = 'stunnel peername lookup failed -> '.$sock_peername;
          return false;
        }
//        else
//        {
//          $stat_msg = 'stunnel peername lookup failed -> '.$sock_peername;
//          $stat_msg = 'connected peername -> '.$sock_peername;
//        }
        
      }
//      else
//        $stat_msg = 'connected peername -> '.$sock_peername;

      if ($_self_inuse) //inuse, not library mode
        $this->sock_peername = $sock_peername;

      return $_sock;
    }
    else if ($config['socket_type'] == SOCK_DGRAM)
    {
      $stat_msg = 'Operation SKIPPED';
      return $sock;
    }
    else
    {
      $stat_msg = 'Operation not supported';
      return false;
    }
  }

  //stunnel utility support options for stream dev
  /**
   * filter stunnel log information
   *
   * parse X last lines from given stunnel log matching filter rules
   *
   * NB, is really a sonic_stream_stunnel_info() wrapper
   *
   * @param string $stl_file full location to stunnel log file
   * @param array &$server_client array with returned information requested by filter
   * @param integer $tail_lines total amount of _most recently entered_ log lines to parse 
   * @param array $filter option flags, (integer) time, (string) server
   * @return boolean true on success and false on failure
  */
  function sonic_stream_stunnel_info_log($stl_file, &$server_client, $tail_lines = 5, $filter = array('time' => 0, 'server' => ''))
  {
    if (file_exists($stl_file));
    else
      return false;

    $tail_lines += 0;

    $st_out = file_get_contents($stl_file);

    if ($tail_lines > 0);
    else
    {
      if (SONIC_stream::sonic_stream_stunnel_info($st_out, $server_client, $filter))
        return $st_out;
      else
        return false;
    }

    $st_out_a = preg_split('/\n/', $st_out);

    $st_out_a_so = sizeof($st_out_a);

    if ($st_out_a_so > $tail_lines);
    else
    {
      unset($st_out_a);
      if (SONIC_stream::sonic_stream_stunnel_info($st_out, $server_client, $filter))
        return $st_out;
      else
        return false;
    }

    $st_out = '';
    for ($i = (($st_out_a_so - 1) - $tail_lines); $i < $st_out_a_so; $i++)
    {
      $st_out .= $st_out_a[$i]."\n"; 
    }

    unset($st_out_a);
    if (SONIC_stream::sonic_stream_stunnel_info($st_out, $server_client, $filter))
      return $st_out;
    else
      return false;
  }

//eg. log format
// 2008.08.09 14:36:52 LOG5[17602:3085847440]: stunnel3 accepted connection from 192.168.0.1:48788
// 2008.08.09 14:36:52 LOG5[17602:3085847440]: stunnel3 connected remote server from 127.0.0.1:40160
  /**
   * filter stunnel log information
   *
   * parse given stunnel log output matching filter rules
   *
   * @param string $stl_out to stunnel log data
   * @param array &$server_client array with returned information requested by filter
   * @param integer $tail_lines total amount of _most recently entered_ log lines to parse
   * @param array $filter option flags, (integer) time, (string) server
   * @return boolean true on success and false on failure
  */
  function sonic_stream_stunnel_info($st_out, &$server_client, $filter = array('time' => 0, 'server' => ''))
  {
    if (gettype($server_client) == array());
    else
      $server_client = array();

    $server_client_so = sizeof($server_client);

    $stl_pat = '/^[0-9]{4,4}\.[0-9]{2,2}\.[0-9]{2,2}\s'
		.'([0-9]{2,2}:[0-9]{2,2}:[0-9]{2,2})\s'
		.'LOG5\[([0-9]+:[0-9]+)\]:\s+stunnel3\s'
		.'((?:accepted\sconnection\sfrom|connected\sremote\sserver\sfrom))\s+'
		.'(.+)\s*$/xm';

    if (preg_match_all($stl_pat, $st_out, $matches));
    else
      return false;

   $matches_row_so = sizeof($matches[0]);

   if (gettype($filter['server']) == 'string' && $filter['server'])
   {
   }
   else
     $filter_server = false;

   if (gettype($filter['time']) == 'boolean' && $filter['time'])
   {
     $filter['time'] = time();
   }
   else if ($filter['time'] + 0 > 0)
   {
     $filter['time'] += 0;
   }
   else
     $filter_time = false;

   $clients = array();
   $servers = array();
   for ($i = 0; $i < $matches_row_so; $i++)
   {
     if ($filter_time && $filter['time'] == $matches[2][$i]) {}
     else if ($filter_time)
       continue;

     if ($matches[3][$i] == 'accepted connection from')
     {
       $clients[$matches[2][$i]] = $matches[4][$i];
     }
     else if ($matches[3][$i] == 'connected remote server from')
     {
       if ($filter_server && $filter['server'] == $matches[4][$i]) {}
       else if ($filter_server)
         continue;

       $servers[$matches[2][$i]] = $matches[4][$i];
     }
   }

   $servers_n_clients_ia = array_unique(array_merge(array_keys($clients), array_keys($servers)));

   foreach ($servers_n_clients_ia as $servers_n_clients_i)
   {
     if ($clients[$servers_n_clients_i] && $servers[$servers_n_clients_i])
     {
       $server_client[$servers[$servers_n_clients_i]] = $clients[$servers_n_clients_i];
     }
     else if ($servers[$servers_n_clients_i])
     {
       $server_client[$servers[$servers_n_clients_i]] = NULL;
     }
   }

   return true;
  }

  /**
   * stunnel_getpeername()
   *
   * map local peername with stunnel index for real connected client peername
   *
   * @param string &$peername host:port format of client peername on socket
   * @param array $filter flag options, (integer) time, (string) server
   * @param array $config third party support (library mode) override, array of type see stream constructor
   * @param resource $sock third party support (library mode) override, see PHP sockets
   * @return boolean boolean see socket_getpeername()
  */
  function sonic_stream_stunnel_getpeername(&$peername, $filter = array('time' => 0, 'server' => ''), $config = NULL, $sock = NULL)
  {
    $peername = NULL;

    if ($this->config == NULL && $this->$sock == NULL)
    {
      $_self_inuse = false;
    }
    else //inuse, not library mode
    { 
      if ($this->stunnel_peername)
      {
        $peername = &$this->stunnel_peername;
        return true;
      }
      $_self_inuse = true;
    }

    if ($config) {}                       //3rd party compat
    else if ($this->config)               //sonic daemon inherit preload
    { 
      $config = $this->config;
    }
    else
      return false;

    if ($sock) {}                         //3rd party compat
    else if ($this->sock)                 //sonic daemon inherit preload
    { 
      $sock = &$this->sock;
    }
    else if ($config['socket'])           //despro search
    { 
      $sock = &$config['socket'];
    }
    else
      return false;

    $stream_config = $config['stream_config'];

    if ($st_log_f = $stream_config['stunnel_log'])
      $st_log_fe = @file_exists($st_log_f);
    $st_usec = $stream_config['stunnel_usec'] + 0;
    if ($st_usec > 0);
    else
      $st_usec = 550;
    if ($stream_config['stunnel_usec_wait_time'] == NULL)
      $st_usec_max = $st_usec * 3;
    else
      $st_usec_max = $stream_config['stunnel_usec_wait_time'] + 0;
    $st_usec_count = 0;
    $st_log_lines = $stream_config['stunnel_log_lines'] + 0;
    if ($st_log_lines > 0);
    else
      $st_log_lines = 10;
    $st_force = $stream_config['stunnel_force'];

    if ($st_force OR $st_log_fe)
    { 
      if ($st_force)
      { 
        if ($st_log_fe) {}
        else if ($st_log_f)
        { 
          return false;
//          die('forcing stunnel but logfile '.$st_log_f.' does not exists'."\n");
        }
        else
          return false;
//          die('forcing stunnel but no logfile given'."\n");
      }

      if (socket_getpeername($sock, $addr_peer, $port_peer))
      {
        if ($filter['server'] == '')
          $filter['server'] = '$addr_peer:$port_peer';
      }
      else
        return false;

      do
      { 
        if ($st_usec_max > 0 && $st_usec_count > $st_usec_max)
        { 
//          echo "timeout\n";
          break;
        }

        usleep($st_usec); //wait for 'connected remote server'
        $st_usec_count += $st_usec;

        $st_log = $this->sonic_stream_stunnel_info_log($st_log_f, $sc, $st_log_lines, $filter);

        if ($sc["$addr_peer:$port_peer"])
          break;

        if ($this->rotate_shutdown)
          break;

//        echo "waiting for stunnel logs to ketchup\n";
      }
      while (true);

      if ($st_pn = $sc["$addr_peer:$port_peer"])
      {
        $peername = $st_pn;

//        echo 'waited '.$st_usec_count.' usecs to fetch peer '.$st_pn.' from stunnel'."\n";
        if ($_self_inuse) //inuse, not library mode
          $this->stunnel_peername = &$peername;

        return true;
      }
    }
    return false;
  }

  function __toString()
  {
    return var_export($this, true);
  }
}

?>
